#ifndef __TTHREAD_HH
#define __TTHREAD_HH
//
//  Project   : ThreadPool
//  File      : TThread.hh
//  Author    : Ronald Kriemann
//  Purpose   : baseclass for a thread-able class
//
// arch-tag: d09c570a-520a-48ce-b612-a813b50e87b4
//

#include <stdio.h>
#include <pthread.h>

#include "types.hh"
#include "TSLL.hh"

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
//
// baseclass for all threaded classes
// - defines basic interface
//
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

class TThread
{
protected:
    // thread-specific things
    pthread_t       _thread_id;
    pthread_attr_t  _thread_attr;

    // is the thread running or not
    bool            _running;

    // no of thread
    int             _thread_no;
    
public:
    ////////////////////////////////////////////
    //
    // constructor and destructor
    //
    
    TThread ( int thread_no = -1 );

    virtual ~TThread ();

    ////////////////////////////////////////////
    //
    // access local data
    //

    int thread_no () const { return _thread_no; }
    int proc_no   () const { return _thread_no; }

    void set_thread_no ( int n );

    // compare if given proc-no is local one
    bool on_proc ( int p ) const { return ((p == -1) || (_thread_no == -1) || (p == _thread_no)); }

    // resets running-status (used in _run_proc, see TThread.cc)
    void reset_running () { _running = false; }
    
    ////////////////////////////////////////////
    //
    // user-interface
    //
    
    // actually method the thread executes
    virtual void run () = 0;

    // start (stop) thread
    virtual void start ( bool detached = false,
                         bool sscope   = false ) { create( detached, sscope ); }
    virtual void stop  () { cancel(); }

    ////////////////////////////////////////////
    //
    // thread management
    //

    // create thread (actually start it)
    void create ( bool detached = false, bool sscope = false );

    // detach thread
    void detach ();
    
    // synchronise with thread (wait until finished)
    void join   ();

    // terminate thread
    void exit   ();

    // request cancellation of thread
    void cancel ();
    
    // send signal signo to thread
    void kill   ( int signo );

    // put thread to sleep (milli + nano seconds)
    void sleep  ( long m_sec, long nano_sec );
};



////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
//
// wrapper for pthread_mutex
//
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

class TMutex
{
    // friends
    friend class TCondition;
    
protected:
    // the mutex itself and the mutex-attr
    pthread_mutex_t      _mutex;
    pthread_mutexattr_t  _mutex_attr;

public:
    /////////////////////////////////////////////////
    //
    // constructor and destructor
    //

    TMutex ()
    {
        pthread_mutexattr_init( & _mutex_attr );
        pthread_mutex_init( & _mutex, & _mutex_attr );
    }

    ~TMutex ()
    {
        pthread_mutex_destroy( & _mutex );
        pthread_mutexattr_destroy( & _mutex_attr );
    }

    /////////////////////////////////////////////////
    //
    // usual behavior of a mutex
    //

    // lock and unlock mutex (return 0 on success)
    int lock    () { return pthread_mutex_lock(   & _mutex ); }
    int unlock  () { return pthread_mutex_unlock( & _mutex ); }

    // return true if mutex is locked, otherwise false
    bool is_locked ()
    {
        if ( pthread_mutex_trylock( & _mutex ) != 0 )
            return true;
        else
        {
            unlock();
            return false;
        }// else
    }
};



////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
//
// class for a condition variable
//
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

class TCondition
{
protected:
    // our condition variable
    pthread_cond_t      _cond;

public:
    /////////////////////////////////////////////////
    //
    // constructor and destructor
    //

    TCondition  () { pthread_cond_init(    & _cond, NULL ); }
    ~TCondition () { pthread_cond_destroy( & _cond ); }

    /////////////////////////////////////////////////
    //
    // condition variable related methods
    //

    void wait      ( TMutex & m ) { m.lock(); pthread_cond_wait( & _cond, & m._mutex ); m.unlock(); }
    void signal    ( TMutex & m ) { m.lock(); pthread_cond_signal(    & _cond );        m.unlock(); }
    void broadcast ( TMutex & m ) { m.lock(); pthread_cond_broadcast( & _cond );        m.unlock(); }
    void signal    ()             { pthread_cond_signal(    & _cond ); }
    void broadcast ()             { pthread_cond_broadcast( & _cond ); }
};

#endif  // __TTHREAD_HH
